#!/usr/bin/perl

use strict;
use Getopt::Long;
use POSIX;
use File::Basename;
use File::Path;

our $version = "3.0";

our $log_file = "/var/log/opdb-dump.log";
our $log_file_error = "/var/log/opdb-dump-errors.log";
our $backup_dir = "/var/tmp/opmondb";
our $mysql = "/usr/bin/mysql --max_allowed_packet=512M";
our $mysqldump = "/usr/bin/mysqldump -q -f --max_allowed_packet=512M --skip-add-locks --skip-comments --skip-disable-keys --single-transaction -c --insert-ignore --quick";
our $mysqladmin = "/usr/bin/mysqladmin";
our ($filename, $newfilename);
our $conf_file = "/usr/local/opmon/etc/db.php";
our $clean_old_backup = "/usr/bin/find $backup_dir -mtime +30 -delete >/dev/null 2>/dev/null";
our %date = _gettime();

# ----- Change Log Process ----- #
our $nohup = "/usr/bin/nohup";
our $php = "/usr/bin/php";
our $clean_db = "/usr/local/opmon/utils/clean_db_opmon4.php > /var/log/opmon-cleandb-optimize.log";
our $clean_db_event_log = "/usr/local/opmon/utils/clean_db_event_log.php > /var/log/opmon-cleandb.log";
# ------------------------------ #

our ( $dbuser, $dbpass, $dbhost, @dblist, @host_ids, @tablelist,$opt_recover,$opt_help,$opt_version, $opt_export );
our ( $opt_full_recover, $opt_full_export, $table_name_tmp, $opt_disaster_recovery );
our @ignoredDBs = ( "information_schema", "backup", "test", "traffic");
our @tablesWithDrop = ( "host_state_change", "service_state_change", "dashboards", "dashboards_resources", "scat_activities", "scat_cf_definitions", "scat_cf_values", "scat_contacts", "scat_itens", "scat_log", "scat_sazdays", "scat_slas", "scats", "interface_shortcut_items", "interface_shortcut_names", "interface_user_favorites", "eventguard_rules", "timecuts", "shared_resources", "host_comment", "service_comment" );

our $error_count=0;

our $usage = '
Usage: '. $0 .' [OPTION]...
Export/Import OpMon database

General Options:
	-r <dbname> - Import  database from /var/tmp/opmondb to mysql
	-e <dbname> - Export  database from mysql to /var/tmp/opmondb
	-R - Recover all databases from /var/tmp/opmondb to mysql
	-E - Export  all databases from mysql to /var/tmp/opmondb
	-D - Disaster recovery - Import needed system tables for quick opmon start
';

sub _main {
	_logger("::: OpMon Database Export/Import tool Starting :::");
	
#		'r:s'	=> \$opt_recover,
#		'e:s'	=> \$opt_export,
#		'h|help'			=> \$opt_help,
#		'v|version'		=> \$opt_version,
#		'R' => \$opt_full_recover,
#		
	print "Logging on file $log_file\n";
	# get command line options
	getOptions();
	
	# If disaser recovery mode set, recover opcfg and special opmon4 tables
	if ($opt_disaster_recovery) {
		$opt_recover="opcfg";
		_logger(" -> Starting database disaster recovery",1);
	}

	# Creates backup dir if it doest exist 
	unless (-e $backup_dir) {
		# Create backup dir
		mkpath($backup_dir);
	} 

	# Read database config
	_read_conf();

	if ($opt_recover) {
		_logger(" -> Starting database $opt_recover recover",1);
		my $db_dir_recover=$backup_dir."/".$opt_recover;
		unless (-e $db_dir_recover) {
			_logger(" -> Dump for $opt_recover does not exist, cant recover it",1);
			exit(0);
		}
		_importfull($opt_recover, $opt_disaster_recovery);
		if ($error_count>0) {
			_logger(" -> Database $opt_recover recover with errors, check log $log_file",1);
		}
		else {
			_logger(" -> Database $opt_recover recover done",1);
		}
		exit(0);
	}
	if ($opt_full_recover) {
		_logger(" -> Starting full database recover",1);
		_importfull();
		if ($error_count>0) {
			_logger(" -> Database recover with errors, check log $log_file",1);
		}
		else {
			_logger(" -> Database recover done",1);
		}
		exit(0);
	}
	if ($opt_export) {
		_logger(" -> Starting database $opt_export export",1);
		_backupfull($opt_export);
		if ($error_count>0) {
			_logger(" -> Database $opt_export export with errors, check log $log_file",1);
		}
		else {
			_logger(" -> Database $opt_export export done",1);
		}
		exit(0);
	}
	if ($opt_full_export) {
		_logger("Starting all databases export",1);
		
		_logger(" -> Clean event_log table");
		system("$php -q $clean_db_event_log >/dev/null 2>/dev/null");
		
		# Clean and optimize database before dumping on the first sunday of the month
		my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
		if ($mday>=1 and $mday<=7) {
			if ($wday==0) {
				_logger(" -> First of the month, running CleanDB Process");
				system("time $php -q $clean_db 2>/dev/null");
			}
		}
		
		# Perform a full backup
		#_slave("stop");	
		_backupfull();
		#_slave("start");
		
		# Clean old SQL dumps
		system($clean_old_backup);
		
		if ($error_count>0) {
			_logger(" -> OpMon Database Backup Finished with errors, check log $log_file",1);
		}
		else {
			_logger("::: OpMon Database Backup Finished :::");
		}
	}
	else {
		_logger("-> No options selceted, nothing done");
		print $usage;
	}
	exit(0);
}

sub get_opmondb_dirs {
	my @array_tmp=`/bin/ls -p $backup_dir | /bin/grep "/" | /usr/bin/tr -d "/"`;
	@dblist=();
	push @dblist, "opcfg";
	foreach (@array_tmp) {
		my $tmp=$_;
		chomp($tmp);
		if ($tmp ne "opcfg") {
			push @dblist, $tmp if (!_in_array($tmp));
		}
	}
	@array_tmp=();
}

sub _gettime {
   my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(time);
   $mon++;
   $year+=1900;

   my %self = (
      'wday'	=>	$wday,
      'year'	=>	$year,
      'mon'	=>	$mon,
      'mday'	=>	$mday,
      'hour'	=>	$hour,
      'min'	=>	$min,
      'sec'	=>	$sec
   );
   return %self;
}

sub _logger {
	my ($msg, $show) = @_;
	
	my %date = _gettime();
	open(_LOG_, ">>$log_file");
	if ($show==1) {
		print "$date{year}/$date{mon}/$date{mday} $date{hour}:$date{min}:$date{sec} - $msg\n";
	}
	print _LOG_ "$date{year}/$date{mon}/$date{mday} $date{hour}:$date{min}:$date{sec} - $msg\n";
	close(_LOG_);
	
	# Remove old log file if is bigger then 10MB
	my @log_stat = stat($log_file);
	if ($#log_stat>=7) {
		unlink($log_file) if ($log_stat[7] >= 10485760);
	}
}

sub _logger_dump_errors {
	my ($msg, $show) = @_;
	
	my %date = _gettime();
	open(_LOG_, ">>$log_file_error");
	if ($show==1) {
		print "$date{year}/$date{mon}/$date{mday} $date{hour}:$date{min}:$date{sec} - $msg\n";

	}
	print _LOG_ "$date{year}/$date{mon}/$date{mday} $date{hour}:$date{min}:$date{sec} - $msg\n";
	close(_LOG_);
	
	# Remove old log file if is bigger then 10MB
	my @log_stat = stat($log_file_error);
	if ($#log_stat>=7) {
		my @log_stat = stat($log_file_error);
		unlink($log_file_error) if ($log_stat[7] >= 10485760);
	}
}

sub _importfull() {
	my ($dbname_to_recover, $disater_recovery)=@_;
	my $ok=0;
	if ($dbname_to_recover) {
		$#dblist=-1;
		push @dblist, $dbname_to_recover;
	}
	else {
		get_opmondb_dirs();
	}
	foreach(@dblist) {
		my $dbname=$_;
		system("$mysqladmin -u $dbuser -h $dbhost --password=$dbpass create $dbname > /dev/null 2>/dev/null");
		# Create directory for tables dump
		my $backup_dir_tmp=$backup_dir."/".$dbname."/*.sql";
		my @sql_files=();
		@sql_files=`ls -1 $backup_dir_tmp`;
		foreach(sort(@sql_files)) {
			my $sql_file=$_;
			chomp($sql_file);
			#_logger(" $dbname -->\t$sql_file");
			$ok=system("$mysql -f -u $dbuser -h $dbhost --password=$dbpass $dbname < $sql_file 2>/dev/null");
			if ($ok==0) {
				_logger(" $dbname <--\t$sql_file OK");
			}
			else {
				$error_count=$error_count+1;
				_logger(" $dbname <--\t$sql_file ERROR");
			}
		}
	}
	if ($disater_recovery) {
		foreach(sort(@tablesWithDrop)) {
			my $sql_file=$_;
			chomp($sql_file);
			$sql_file=$backup_dir."/opmon4/".$sql_file.".sql";
			$ok=system("$mysql -f -u $dbuser -h $dbhost --password=$dbpass opmon4 < $sql_file 2>/dev/null");
			if ($ok==0) {
				_logger(" opmon4 <--\t$sql_file OK");
			}
			else {
				$error_count=$error_count+1;
				_logger(" opmon4 <--\t$sql_file ERROR");
			}
		}
		# Create host_state_change and service_state_change tables
		_get_host_ids();
		foreach (@host_ids) {
			my $tmp = "host_state_change_".$_;
			_logger(" opmon4 <--\tcreating table $tmp OK");
			my $host_table="CREATE TABLE opmon4.host_state_change_".$_." LIKE opmon4.host_state_change";
			_run_query($host_table);
			
			my $tmp = "service_state_change_".$_;
			_logger(" opmon4 <--\tcreating table $tmp OK");
			my $service_table="CREATE TABLE opmon4.service_state_change_".$_." LIKE opmon4.service_state_change";
			_run_query($service_table);
		}	
	}
}

sub _backupfull {
	my ($dbname_to_export)=@_;
	my $ok=0;
	
	# Get database backup list
	_get_databases_list(); 
	
	if ($dbname_to_export) {
		unless (grep {$_ eq $dbname_to_export} @dblist) {
			_logger(" -> Database $dbname_to_export does not exist to be exported",1);
			exit(0);
		}
		$#dblist=-1;
		push @dblist, $dbname_to_export;
		# Clean old .sql from old export format
		my $old_sql_files_tmp="/bin/rm -f ".$backup_dir."/*.sql";
		system($old_sql_files_tmp);		
	}
	else {
		_logger(" -> Dumping all databases",1);
		# Clean old .sql from old export format
		my $old_sql_files_tmp="/bin/rm -f ".$backup_dir."/*.sql";
		system($old_sql_files_tmp);
	}

	# Dump each database in a diffrent file  
  foreach(sort(@dblist)) {
  	my $dbname=$_;
    # Create directory for tables dump
    my $backup_dir_tmp=$backup_dir."/".$dbname;
    mkdir($backup_dir_tmp); 
		_get_tables_list($dbname); 
    foreach(sort(@tablelist)) {
    	my $table_name=$_;
			$table_name_tmp=$table_name.".sql-tmp";
			if ($dbname eq "opcfg" or $dbname eq "mysql") {
      	$ok=system("$mysqldump -u $dbuser -h $dbhost --password=$dbpass --add-drop-table $dbname $table_name > $backup_dir_tmp/$table_name_tmp 2>/dev/null");
      }
      else {
				if ( grep( /^$table_name/, @tablesWithDrop ) ) {
      		$ok=system("$mysqldump -u $dbuser -h $dbhost --password=$dbpass --add-drop-table $dbname $table_name > $backup_dir_tmp/$table_name_tmp 2>/dev/null");
      	}
      	else {
      		$ok=system("$mysqldump -u $dbuser -h $dbhost --password=$dbpass --skip-add-drop-table $dbname $table_name > $backup_dir_tmp/$table_name_tmp 2>/dev/null");
      	}
    	}
			if ($ok==0) {
				_logger(" $dbname -->\t$table_name OK");
			}
			else {
				$error_count=$error_count+1;
				_logger(" $dbname -->\t$table_name ERROR");
			}
    	# Check if file has 0 bytes, if so, do not delete old one
    	$filename=$backup_dir_tmp."/".$table_name_tmp;
			$newfilename=$backup_dir_tmp."/".$table_name.".sql";
    	our ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat($filename);
    	if ($size eq "") {
    		# File does not exits
    		_logger_dump_errors(" $dbname -->\t$table_name - did not export");
			}
			else {
				if ($size > 0) {
					# File exported OK, remove old and rename tmp file
					unlink($newfilename);
					rename($filename,$newfilename);
				}
				else {
					# print "File is zero sized\n";
					_logger_dump_errors(" $dbname -->\t$table_name - 0 size export file, probably corrupted table");
					unlink($filename);
				}
			}
		}
	}
}

sub _read_conf {

   _logger(" -> Reading database config");

   open( FILE, "$conf_file" );
   while(<FILE>) {
      $dbuser = $1 if (m/DBUSER\s*=\s*[\'\"]*([^\'\"]*)[\'\"]*;/);
      $dbpass = $1 if (m/DBPASS\s*=\s*[\'\"]*([^\'\"]*)[\'\"]*;/);
      $dbhost = $1 if (m/DBHOST\s*=\s*[\'\"]*([^\'\"]*)[\'\"]*;/);
   }
   close(FILE);

}

sub _get_databases_list {
   my @list = `$mysql -u $dbuser -h $dbhost --password=$dbpass <<EOF | grep -v Database
show databases;
quit
EOF`;

	push @dblist, "opcfg";
	foreach(@list) {
		chomp;
		push @dblist, $_ if (!_in_array($_));
	}
}

sub _run_query {
	my ($sql)=@_;
	my @list = `$mysql -u $dbuser -h $dbhost --password=$dbpass <<EOF | grep -v Database
$sql;
quit
EOF`;

}

sub _get_host_ids {
	my @list = `$mysql -u $dbuser -h $dbhost --password=$dbpass <<EOF | grep -v host_id
select host_id from opcfg.nagios_hosts order by 1;
quit
EOF`;

	foreach(@list) {
		chomp;
		push @host_ids, $_ if (!_in_array($_));
	}
}

sub _get_tables_list {
	my ($dbname)=@_;
	 
	my @list = `$mysql -u $dbuser -h $dbhost --password=$dbpass --database=$dbname <<EOF | grep -v Tables_in_
show tables;
quit
EOF`;

	@tablelist=();
	foreach(@list) {
		chomp;
		push @tablelist, $_ if (!_in_array($_));
	}
}

sub _in_array {
	my $value = shift;
	return 0 if ($value eq $opt_export);
	foreach(@ignoredDBs) {
		return 1 if m/$value/;
	}
	return 0;
}

sub _slave {
	my $param = shift;
	 _logger(" -> Slave $param");
	`$mysql -h $dbhost -u $dbuser --password="$dbpass" <<SQL
slave $param;
exit
SQL`;
}

sub getOptions {  #command line options
	Getopt::Long::Configure('bundling');
  GetOptions(
		'r:s'	=> \$opt_recover,
		'e:s'	=> \$opt_export,
		'h|help|?'			=> \$opt_help,
		'v|version'		=> \$opt_version,
		'R' => \$opt_full_recover,
		'E' => \$opt_full_export,
		'D' => \$opt_disaster_recovery,
	)or do {
			exit();
		};

	if ($opt_version) {
  	print_version($version);
	}
	if ($opt_help) {
  	print $usage;
  	exit(0);
	}
}

&_main;
